/*
 *  plex86: run multiple x86 operating systems concurrently
 *  Copyright (C) 1999-2000  The plex86 developers team
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "plex86.h"
#include "user.h"
#include "plugin.h"


/************************************************************************/
/* Declarations                                                         */
/************************************************************************/

static void abort_handler (int i);
callback_command_t *callback_command_list;



/************************************************************************/
/* Structures / Variables                                               */
/************************************************************************/

/* Global data */

config_info_t vm_conf;



/************************************************************************/
/* Main code                                                            */
/************************************************************************/

  int
main (int argc, char *argv[])
{
    int   config_fd, line_index, i;
    char  line[256], sxname[10], config_file_name[256];
    char  c;
    struct sigaction sg_act;



    /* some inits */

    strcpy (config_file_name, "plex86.conf");
    vm_conf.max_memory    =  0;
    vm_conf.text_address  = -1;
    vm_conf.stack_address = -1;
    vm_conf.verbose       =  0;
    vm_conf.dump_vm       =  0;
    vm_conf.exit_wait     =  0;
    vm_conf.syntax        =  SX_NONE;
    plugin_startup ();



    /* parsing command line for arguments */

    fprintf (stderr, "Processing command-line options\n");

/* +++ fix weird use of argv[i][3] below */

    for (i = 0; i < argc; i++)
    {
	if (argv[i][0] != '-')
	    continue;

	switch (argv[i][1])
	{
	case 'h':
	    printf ("plex86 user-level monitor version " VERSION "\n");
            printf ("Usage:  user -fhmsv\n");
            printf ("Options:\n");
	    printf ("-f ... Config File (fully qualified path)\n");
	    printf ("-h ... Print this helpful help\n");
	    printf ("-m ... Megs of memory for VM\n");
	    printf ("-s ... Select disassembly syntax (\"intel\" or \"att\")\n");
            printf ("-w ... Wait <n> seconds before exiting when a panic occurs\n");
	    printf ("-v ... Print debug messages (verbose)\n");
	    exit (0);
	    break;

	case 'm':
	    vm_conf.max_memory = atoi (&argv[i][3]);
	    break;

        case 'w':
            vm_conf.exit_wait = atoi (&argv[i][3]);
            break;

	case 'f':
	    strncpy (config_file_name, &argv[i][3], 255);
	    break;

#if 0
#if USE_LOADER
	case 'l':
            {
            extern char *special_loader_fname;
            special_loader_fname = strdup(argv[i+1]);
	    break;
            }
#endif
#endif

	case 's':
	    strncpy (sxname, &argv[i][3], 10);
	    if (!strcmp (sxname, "intel"))
		vm_conf.syntax = SX_INTEL;
	    else if (!strcmp (sxname, "att"))
		vm_conf.syntax = SX_ATT;
	    break;

	case 'v':
	    vm_conf.verbose = 1;
	    break;

	default:
	    fprintf (stderr, "Unknown option: %s\n", argv[i]);
	    break;
	}
    }

    vm_open();


    /* read config file */

    fprintf (stderr, "Processing configuration script %s\n", config_file_name);

    config_fd = open (config_file_name, O_RDONLY);
    if (config_fd == -1)
    {
	char  error_msg[100];

	sprintf (error_msg, "error while opening %s", config_file_name);
	perror (error_msg);
	exit (1);
    }

    line_index = 0;		/* states: 0 = reading, 1 = ignoring */

/* +++ Ack!  We're not doing any bounds check here.  This kind */
/* of file IO is not efficient either. */

    while (read (config_fd, &c, 1))
    {
        /* If continuation character, then keep going. */
        if (c == '\\') {
          (void) read (config_fd, &c, 1); /* eat the '\n' */
          continue;
          }

	line[line_index++] = c;
	if (c == '\n')
	{
	    line[line_index - 1] = '\0';
	    line_index = 0;
	    if ((line[0] == '#') || (line[0] == '\n'))
		continue;	/* is a comment or a blank line; */

	    while ((line[line_index] != ' ') && (line[line_index] != '='))
		line_index++;
	    if (!strncmp (line, "memory", line_index))
	    {
		if (vm_conf.max_memory != 0)
		{
		    line_index = 0;
		    continue;
		}		/* command line overwrites config file settings; */

		while ((line[line_index] == ' ') || (line[line_index] == '='))
		    line_index++;
		vm_conf.max_memory = atoi (&line[line_index]);
                vm_init_memory(vm_conf.max_memory);
	    }
	    else if (!strncmp (line, "db_syntax", line_index))
	    {
		if (vm_conf.syntax != SX_NONE)
		{
		    line_index = 0;
		    continue;
		}

		while ((line[line_index] == ' ') || (line[line_index] == '='))
		    line_index++;
		strcpy (sxname, &line[line_index]);

		if (!strcmp (sxname, "intel"))
		    vm_conf.syntax = SX_INTEL;
		else if (!strcmp (sxname, "at&t"))
		    vm_conf.syntax = SX_ATT;
	    }
	    else if (!strncmp (line, "prescan_depth", line_index))
	    {
                if (vm_conf.prescanDepth != 0)
                {
                    line_index = 0;
                    continue;
                }               /* command line overwrites config file settings; */

                while ((line[line_index] == ' ') || (line[line_index] == '='))
                    line_index++;
                vm_conf.prescanDepth = atoi (&line[line_index]);
                vm_init_prescan_depth(vm_conf.prescanDepth);
	    }
            else if (!strncmp (line, "exit_wait", line_index))
            {
                if (vm_conf.exit_wait != 0)
                {
                    line_index = 0;
                    continue;
                } /* command line overwrites config file settings; */
 
                while ((line[line_index] == ' ') || (line[line_index] == '='))
                    line_index++;
                vm_conf.exit_wait = atoi (&line[line_index]);
            }
	    else if (!strncmp (line, "plugin", line_index))
	    {
		int   parm_index;
		char *plugin_name, *plugin_args;

		while ((line[line_index] == ' ') || (line[line_index] == '='))
		    line_index++;

		parm_index = line_index;
		while ((line[parm_index] != ' ') && (line[parm_index] != '\0'))
		    parm_index++;
		if (line[parm_index] != '\0')
		    line[parm_index++] = '\0';

		plugin_name = strcpy (malloc (strlen (&line[line_index]) + 1), &line[line_index]);
		plugin_args = strcpy (malloc (strlen (&line[parm_index]) + 1), &line[parm_index]);

		plugin_load (plugin_name, plugin_args);
	    }
	    else if (!strncmp (line, "load-rom", line_index)) {
               Bit32u address;
               char *fname_ptr, *address_ptr;

              while ( line[line_index] == ' ' )
                line_index++;
              if ( strncmp(&line[line_index], "file=", strlen("file=")) ) {
                fprintf(stderr, "load-rom directive syntax error:"
                        " expecting 'file='\n");
                exit(0);
                }
              line_index += strlen("file=");
              fname_ptr = &line[line_index];
              while ( line[line_index] && (line[line_index] != ' ') )
                line_index++;
              while ( line[line_index] == ' ' )
                line_index++;
              if ( strncmp(&line[line_index], "address=", strlen("address=")) ) {
                fprintf(stderr, "load-rom directive syntax error:"
                        " expecting 'address='\n");
                exit(0);
                }
              line[line_index-1] = 0; /* terminate filename string */
              line_index += strlen("address=");
              address_ptr = &line[line_index];
              address = strtoul(address_ptr, NULL, 0);
              if (vm_load_rom(fname_ptr, address)) {
                fprintf(stderr, "load-rom '%s' @ 0x%x failed.\n",
                        fname_ptr, address);
                exit(0);
                }
	      }
            else {
              /* Command is not one of the predefined ones.  Check */
              /* list of registered commands from the plugin modules. */
              callback_command_t *command = callback_command_list;
              while (command) {
                if (!strncmp(line, command->command, line_index)) {
                  command->f(&line[line_index+1]);
                  break;
                  }
                command = command->next;
                }
              }
	    line_index = 0;
	}
    }

    close (config_fd);



    /* set default settings, if not specified */

    if (vm_conf.max_memory == 0)
	vm_conf.max_memory = 4;
    if (vm_conf.syntax == SX_NONE)
	vm_conf.syntax = SX_ATT;
    if (vm_conf.text_address == -1)
	vm_conf.text_address = 0;
    if (vm_conf.stack_address == -1)
      vm_conf.stack_address = 0;



    /* setup to catch SIGINT, SIGABRT and SIGQUIT signals so we can clean up */

    memset (&sg_act, 0, sizeof (sg_act));
    sg_act.sa_handler = abort_handler;
    sg_act.sa_flags = SA_RESETHAND;
    sigaction (SIGINT, &sg_act, NULL);
    sigaction (SIGABRT, &sg_act, NULL);
    sigaction (SIGQUIT, &sg_act, NULL);



    /* kickstart the VM */

    vm_kickstart ();

    /* Verify if addresses specified are valid */
    
    if (vm_conf.verbose)
    {
	fprintf (stderr, "Memory:\t\t%d\n",      vm_conf.max_memory);
	fprintf (stderr, "Text address:\t0x%x\n",  vm_conf.text_address);
	fprintf (stderr, "Stack address:\t0x%x\n", vm_conf.stack_address);
    }

    if ( vm_conf.max_memory*1024*1024 < vm_conf.text_address ||
         vm_conf.max_memory*1024*1024 < vm_conf.stack_address )
    {
        fprintf (stderr, "Total memory for VM less than some of the \n");
        fprintf (stderr, "  specified segment addresses, Quiting...\n");
        exit (1);
    }

    /* run the VM */

    fprintf (stderr, "Running VM\n");
#if COSIMULATE
    {
    int   bx_dbg_main(int argc, char *argv[]);
    int   dbg_argc = 1;
    char *dbg_argv[] = { "plex86", NULL };

    /* Pass execution to cosimulation controller */
    bx_dbg_main(dbg_argc, dbg_argv);
    }
#else
    vm_event_loop();
#endif



    /* vm_event_loop() should never return */

    fprintf (stderr, "Fatal: vm_event_loop() returned.  This is a bug. ");
    fprintf (stderr, "Please contact " BUGSMAIL "\n");
    return 1;
}



/************************************************************************/
/* Signal handlers                                                      */
/************************************************************************/

static void
abort_handler (int i)
{
    switch (i)
    {
    case SIGINT:
        printf ("^C pressed. aborting execution\n");
        break;

    case SIGABRT:
    case SIGQUIT:
        printf ("^\\ pressed. dumping VM core and aborting execution\n");
        vm_conf.dump_vm = 1;
        break;

    default:
        printf ("Signal %d caught, aborting execution\n", i);
        break;
    }

    vm_abort ();
    return;
}
